package de.tu_darmstadt.tk.SmartHomeNetworkSim.view.menuBar;

import org.apache.commons.math3.distribution.NormalDistribution;

import de.tu_darmstadt.tk.SmartHomeNetworkSim.control.Controller;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.control.NetworkController;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.control.PacketCaptureController;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.control.SimulationController;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Connection;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.ConnectionPrecision;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Link;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.PacketCollector;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Port;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.PrecisionLink;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.RoomStatus;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.SimulationManager;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.SmartDevice;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.devices.SmartLight;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.devices.SmartLightSensor;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.devices.SmartTemperatureProducer;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.devices.SmartTemperatureSensor;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.distributionHandler.NormalDistributionHandler;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.protocols.MQTT_protocol;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.scheduler.AbstractEvent;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.evaluation.SWCKMeansClustering;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.evaluation.SWCKMeansClusteringWithLabels;

public class LightScenario {

	Controller controller;
	NetworkController networkController;
	public LightScenario(Controller controller) {
		this.controller=controller;
		this.networkController=controller.getNetworkController();
	}
	
	/**
	 * Creates the paper toy example, also simulates it, if run = true
	 * @param run
	 */
	@SuppressWarnings("unchecked")
	public void createSWCExample(boolean run){
		/**
		 * Reset network, etc.
		 */

		long hour = 60 * 60 * 1000;
		long endTime = 24 * hour;
		long noAnomEnd = endTime + 1 * hour;
		long lightEnd = noAnomEnd + 1 * hour;
		long stepLength = hour;
		long currentSimTime=0;
		
		SimulationController sim = controller.getSimulationController();
		if(run) {
			controller.getSimulationController().resetSimulation();
			controller.getNetworkController().deleteNetworkModel();
			/*
			 * Simulate 24 hours
			 */
			sim.setStartTime(0);
			sim.resetSimulation();
			sim.setStepDuration(hour);
			sim.setPrintPackets(false);
			sim.setEndTime(endTime);
		}
		/*
		 * Main networking devices 
		 */
		SmartDevice smartHub = networkController.createSmartDevice("SmartHub", 400, 400, 50);
		
		/*
		 * Links/Networks
		 */
		Link zigbee = new PrecisionLink("ZigBee");
		networkController.addLink(zigbee);
		
		/*
		 * Connect Devices to Links
		 */
		networkController.addLinkToDevice(zigbee, smartHub);
		
		
		/*
		 * Create MQTT Connection
		 */
		Connection mqtt = new ConnectionPrecision();
		mqtt.setName("HomeAutomation (MQTT)");
		networkController.addConnectionToLink(mqtt, zigbee);
		mqtt.setProtocol(new MQTT_protocol());
		Port pBroker = new Port(smartHub, (short)1883);
		pBroker.setStatus(Port.OPEN);
		pBroker.setResponseTime((short)2);
		networkController.addDeviceToConnectionAndProtocol(pBroker, mqtt, 0);
		networkController.addConnection(mqtt);
		mqtt.setLabel((short) 0);
		
		/**
		 * Room
		 */
		String roomName = "Bedroom";
		RoomStatus room = new RoomStatus();
		room.setTemperature(19f);
		/*
		 * Add some MQTT Devices
		 */
		
		/*
		 * Add light Sensor
		 */
		SmartLightSensor lightSensor = new SmartLightSensor(roomName + " LightSensor", room, 1000);
		lightSensor.setBSinfoName("home/" + roomName + "/light");
		networkController.addLinkToDevice(zigbee, lightSensor);
		networkController.moveSmartDevice(lightSensor, 500, 300, 50);
		networkController.addSmartDevice(lightSensor);
		Port pLightSensor = new Port(lightSensor, (short)1883, 15000);
		pLightSensor.setTriggerHandler(new NormalDistributionHandler(15000, 500));
		pLightSensor.setStatus(Port.SENDING);
		pLightSensor.setLastTrigger(-1337L);
		networkController.addDeviceToConnectionAndProtocol(pLightSensor, mqtt,1);
		lightSensor.setLabel((short) 0);
		
		/*
		 * Add a light
		 */
		SmartLight smartLight = new SmartLight(roomName + " Light", room, lightSensor);
		smartLight.setBSinfoName("home/" + roomName + "/light");
		networkController.addLinkToDevice(zigbee, smartLight);
		networkController.moveSmartDevice(smartLight, 500, 500, 50);
		networkController.addSmartDevice(smartLight);
		Port pSmartLight = new Port(smartLight, (short)1883, 15000);
		pSmartLight.setTriggerHandler(new NormalDistributionHandler(15000, 500));
		pSmartLight.setStatus(Port.SENDING);
		pSmartLight.setLastTrigger(-1207L);
		networkController.addDeviceToConnectionAndProtocol(pSmartLight, mqtt,1);
		smartLight.setLabel((short) 0);
		
		/* 
		 * Update visualization 
		 */
		controller.notifyObservers();
		
		/**
		 * Run only if run == true
		 */
		if(!run)return;
		try {
			System.out.println("Check 1");//TODO
			/**
			 * Instances of the packet Sniffers
			 */
			SWCKMeansClusteringWithLabels snifferKNNwithLabels = new SWCKMeansClusteringWithLabels();
			SWCKMeansClustering snifferKNNnoLabels = new SWCKMeansClustering();


			PacketCollector collectorKNNwithLabels = new PacketCollector(snifferKNNwithLabels);
			PacketCollector collectorKNNnoLabels = new PacketCollector(snifferKNNnoLabels);
			//PacketCollector collectorEM = new PacketCollector(snifferEM);
			//PacketCollector collectorHC = new PacketCollector(snifferHC);
			
			/*
			 * Capture both links on all collectors
			 */
			PacketCaptureController captureController = controller.getSimulationController().getPacketCaptureController();
			//captureController.addLinkToCollector(collectorEM, zigbee);
			//captureController.addLinkToCollector(collectorKNN, zigbee);
			captureController.addDeviceToCollector(collectorKNNwithLabels, smartLight);
			captureController.addPacketCollector(collectorKNNwithLabels);
			captureController.addDeviceToCollector(collectorKNNnoLabels, smartLight);
			captureController.addPacketCollector(collectorKNNnoLabels);
			//captureController.addPacketCollector(collectorEM);
			//captureController.addPacketCollector(collectorHC);

			System.out.println("Check 3: created Controller");//TODO
			
			long currentTime = System.currentTimeMillis();
			


			/**
			 * Training events
			 */
			System.out.println("Create training events");
			/**
			 * Light Events (random switch every 30min+-15min
			 */
			NormalDistribution lightDist = new NormalDistribution(hour/10, hour/20);
			long eventTime = 0;
			boolean on = true;
			while(eventTime < noAnomEnd) {
				/**
				 * At least 1 second between events
				 */
				
				final boolean newVal = on;
				SimulationManager.scheduleEvent(new AbstractEvent(eventTime) {
					
					@Override
					public void simulateEvent(long time) {
						smartLight.setTrueStatus(newVal);
						smartLight.setBSval(newVal);
					}
				});
				on = !on;
				eventTime+=Math.max(1000, Math.round(lightDist.sample()));
				
			}
			
			System.out.println("Training:");//TODO
			while(currentSimTime<endTime) {
				sim.getSimulationManager().simulateTimeIntervall(currentSimTime, stepLength);
				currentSimTime+=stepLength;
			}
			
			long new_time = System.currentTimeMillis();
			long elapsed_time = new_time-currentTime;
			System.out.println("Training data generation took: "+elapsed_time+"ms");
			currentTime = new_time;
			
			/*
			 * Start Classifying 
			 */
			collectorKNNwithLabels.setMode(true);
			collectorKNNnoLabels.setMode(true);
			//collectorEM.setMode(true);
			//collectorHC.setMode(true);
			new_time = System.currentTimeMillis();
			elapsed_time = new_time-currentTime;
			System.out.println("Training of algorithm took: "+elapsed_time+"ms");
			currentTime = new_time;
			
			
			
			/*
			 * Simulate/Test 2 hour without anomalies
			 */
			snifferKNNwithLabels.setCurrentScenario("NoAnomalies");
			snifferKNNnoLabels.setCurrentScenario("NoAnomalies");
			System.out.println("Test w/0 anomalies:");//TODO
			sim.getSimulationManager().simulateTimeIntervall(currentSimTime, noAnomEnd-currentSimTime);
			currentSimTime=noAnomEnd;
			
			
			new_time = System.currentTimeMillis();
			elapsed_time = new_time-currentTime;
			System.out.println("Test witout anomalies took: "+elapsed_time+"ms");
			currentTime = new_time;
			
			/**
			 * Light Anomaly 1h
			 * Light != Sensor
			 */
			snifferKNNwithLabels.setCurrentScenario("LightAnomalies");
			snifferKNNnoLabels.setCurrentScenario("LightAnomalies");
			System.out.println("Light Anomaly:");
			eventTime = noAnomEnd;
			on = true;
			smartLight.setTrueStatus(false);
			smartLight.setBSval(true);
			while(eventTime < lightEnd) {
				/**
				 * At least 1 second between events
				 */
				
				final boolean newVal = on;
				SimulationManager.scheduleEvent(new AbstractEvent(eventTime) {
					
					@Override
					public void simulateEvent(long time) {
						smartLight.setTrueStatus(newVal);
						smartLight.setBSval(!newVal);
					}
				});
				on = !on;
				eventTime+=Math.max(1000, Math.round(lightDist.sample()));
				
			}
			smartLight.setLabel((short)-1);
			sim.getSimulationManager().simulateTimeIntervall(currentSimTime, lightEnd-currentSimTime);
			currentSimTime=lightEnd;

			smartLight.setBSval(smartLight.isTrueStatus());
			smartLight.setLabel((short)0);
			
			new_time = System.currentTimeMillis();
			elapsed_time = new_time-currentTime;
			System.out.println("Anomaly generation & classification: "+elapsed_time+"ms");
			currentTime = new_time;
			
			System.out.println("Testing completed");
		} catch(Exception e) {
			System.out.println("WARNING: Testing failed: ");
			e.printStackTrace();
		}
	}
}
